#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
import signal
import time

import fixtures
from pifpaf.drivers import rabbitmq

from oslo_messaging.tests.functional import utils
from oslo_messaging.tests import utils as test_utils


class ConnectedPortMatcher(object):
    def __init__(self, port):
        self.port = port

    def __eq__(self, data):
        return data.get("port") == self.port

    def __repr__(self):
        return "<ConnectedPortMatcher port=%d>" % self.port


class RabbitMQFailoverTests(test_utils.BaseTestCase):
    DRIVERS = [
        "rabbit",
    ]

    def test_failover_scenario(self):
        # NOTE(sileht): run this test only if functional suite run of a driver
        # that use rabbitmq as backend
        self.driver = os.environ.get('TRANSPORT_DRIVER')
        if self.driver not in self.DRIVERS:
            self.skipTest("TRANSPORT_DRIVER is not set to a rabbit driver")

        # NOTE(sileht): Allow only one response at a time, to
        # have only one tcp connection for reply and ensure it will failover
        # correctly
        self.config(heartbeat_timeout_threshold=1,
                    rpc_conn_pool_size=1,
                    kombu_reconnect_delay=0,
                    rabbit_retry_interval=0,
                    rabbit_retry_backoff=0,
                    group='oslo_messaging_rabbit')

        self.pifpaf = self.useFixture(rabbitmq.RabbitMQDriver(cluster=True,
                                                              port=5692))

        self.url = self.pifpaf.env["PIFPAF_URL"]
        self.n1 = self.pifpaf.env["PIFPAF_RABBITMQ_NODENAME1"]
        self.n2 = self.pifpaf.env["PIFPAF_RABBITMQ_NODENAME2"]
        self.n3 = self.pifpaf.env["PIFPAF_RABBITMQ_NODENAME3"]

        # NOTE(gdavoian): additional tweak for pika driver
        if self.driver == "pika":
            self.url = self.url.replace("rabbit", "pika")

        # ensure connections will be establish to the first node
        self.pifpaf.stop_node(self.n2)
        self.pifpaf.stop_node(self.n3)

        self.servers = self.useFixture(utils.RpcServerGroupFixture(
            self.conf, self.url, endpoint=self, names=["server"]))

        # Don't randomize rabbit hosts
        self.useFixture(fixtures.MockPatch(
            'oslo_messaging._drivers.impl_rabbit.random',
            side_effect=lambda x: x))

        # NOTE(sileht): this connects server connections and reply
        # connection to nodename n1
        self.client = self.servers.client(0)
        self.client.ping()
        self._check_ports(self.pifpaf.port)

        # Switch to node n2
        self.pifpaf.start_node(self.n2)
        self.assertEqual("callback done", self.client.kill_and_process())
        self.assertEqual("callback done", self.client.just_process())
        self._check_ports(self.pifpaf.get_port(self.n2))

        # Switch to node n3
        self.pifpaf.start_node(self.n3)
        time.sleep(0.1)
        self.pifpaf.kill_node(self.n2, signal=signal.SIGKILL)
        time.sleep(0.1)
        self.assertEqual("callback done", self.client.just_process())
        self._check_ports(self.pifpaf.get_port(self.n3))

        self.pifpaf.start_node(self.n1)
        time.sleep(0.1)
        self.pifpaf.kill_node(self.n3, signal=signal.SIGKILL)
        time.sleep(0.1)
        self.assertEqual("callback done", self.client.just_process())
        self._check_ports(self.pifpaf.get_port(self.n1))

    def kill_and_process(self, *args, **kargs):
        self.pifpaf.kill_node(self.n1, signal=signal.SIGKILL)
        time.sleep(0.1)
        return "callback done"

    def just_process(self, *args, **kargs):
        return "callback done"

    def _get_log_call_startswith(self, filter):
        return [call for call in self.logger.debug.mock_calls
                if call[1][0].startswith(filter)]

    def _check_ports(self, port):
        getattr(self, '_check_ports_%s_driver' % self.driver)(port)

    def _check_ports_pika_driver(self, port):
        rpc_server = self.servers.servers[0].server
        # FIXME(sileht): Check other connections
        connections = [
            rpc_server.listener._connection
        ]
        for conn in connections:
            self.assertEqual(
                port, conn._impl.socket.getpeername()[1])

    def _check_ports_rabbit_driver(self, port):
        rpc_server = self.servers.servers[0].server
        connection_contexts = [
            # rpc server
            rpc_server.listener._poll_style_listener.conn,
            # rpc client
            self.client.client.transport._driver._get_connection(),
            # rpc client replies waiter
            self.client.client.transport._driver._reply_q_conn,
        ]

        for cctxt in connection_contexts:
            socket = cctxt.connection.channel.connection.sock
            self.assertEqual(port, socket.getpeername()[1])



# Copyright (C) 2015 Cisco Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import kafka
from kafka.common import KafkaError
import mock
from oslo_serialization import jsonutils
import testscenarios
import time

import oslo_messaging
from oslo_messaging._drivers import common as driver_common
from oslo_messaging._drivers import impl_kafka as kafka_driver
from oslo_messaging.tests import utils as test_utils

load_tests = testscenarios.load_tests_apply_scenarios


class TestKafkaDriverLoad(test_utils.BaseTestCase):

    def setUp(self):
        super(TestKafkaDriverLoad, self).setUp()
        self.messaging_conf.transport_driver = 'kafka'

    def test_driver_load(self):
        transport = oslo_messaging.get_transport(self.conf)
        self.assertIsInstance(transport._driver, kafka_driver.KafkaDriver)


class TestKafkaTransportURL(test_utils.BaseTestCase):

    scenarios = [
        ('none', dict(url=None,
                      expected=dict(hostaddrs=['localhost:9092']))),
        ('empty', dict(url='kafka:///',
                       expected=dict(hostaddrs=['localhost:9092']))),
        ('host', dict(url='kafka://127.0.0.1',
                      expected=dict(hostaddrs=['127.0.0.1:9092']))),
        ('port', dict(url='kafka://localhost:1234',
                      expected=dict(hostaddrs=['localhost:1234']))),
        ('two', dict(url='kafka://localhost:1234,localhost2:1234',
                     expected=dict(hostaddrs=['localhost:1234',
                                              'localhost2:1234']))),

    ]

    def setUp(self):
        super(TestKafkaTransportURL, self).setUp()
        self.messaging_conf.transport_driver = 'kafka'

    def test_transport_url(self):
        transport = oslo_messaging.get_transport(self.conf, self.url)
        self.addCleanup(transport.cleanup)
        driver = transport._driver

        conn = driver._get_connection(kafka_driver.PURPOSE_SEND)
        self.assertEqual(self.expected['hostaddrs'], conn.hostaddrs)


class TestKafkaDriver(test_utils.BaseTestCase):
    """Unit Test cases to test the kafka driver
    """

    def setUp(self):
        super(TestKafkaDriver, self).setUp()
        self.messaging_conf.transport_driver = 'kafka'
        transport = oslo_messaging.get_transport(self.conf)
        self.driver = transport._driver

    def test_send(self):
        target = oslo_messaging.Target(topic="topic_test")
        self.assertRaises(NotImplementedError,
                          self.driver.send, target, {}, {})

    def test_send_notification(self):
        target = oslo_messaging.Target(topic="topic_test")

        with mock.patch.object(
                kafka_driver.Connection, 'notify_send') as fake_send:
            self.driver.send_notification(target, {}, {}, None)
            self.assertEqual(1, len(fake_send.mock_calls))

    def test_listen(self):
        target = oslo_messaging.Target(topic="topic_test")
        self.assertRaises(NotImplementedError, self.driver.listen, target,
                          None, None)


class TestKafkaConnection(test_utils.BaseTestCase):

    def setUp(self):
        super(TestKafkaConnection, self).setUp()
        self.messaging_conf.transport_driver = 'kafka'
        transport = oslo_messaging.get_transport(self.conf)
        self.driver = transport._driver

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, '_send')
    def test_notify(self, fake_send, fake_ensure_connection):
        conn = self.driver._get_connection(kafka_driver.PURPOSE_SEND)
        conn.notify_send("fake_topic", {"fake_ctxt": "fake_param"},
                         {"fake_text": "fake_message_1"}, 10)
        self.assertEqual(1, len(fake_send.mock_calls))

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, '_send')
    def test_notify_with_retry(self, fake_send, fake_ensure_connection):
        conn = self.driver._get_connection(kafka_driver.PURPOSE_SEND)
        fake_send.side_effect = KafkaError("fake_exception")
        conn.notify_send("fake_topic", {"fake_ctxt": "fake_param"},
                         {"fake_text": "fake_message_2"}, 10)
        self.assertEqual(10, len(fake_send.mock_calls))

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, '_parse_url')
    def test_consume(self, fake_parse_url, fake_ensure_connection):
        fake_message = {
            "context": {"fake": "fake_context_1"},
            "message": {"fake": "fake_message_1"}}

        conn = kafka_driver.Connection(
            self.conf, '', kafka_driver.PURPOSE_LISTEN)

        conn.consumer = mock.MagicMock()
        conn.consumer.fetch_messages = mock.MagicMock(
            return_value=iter([jsonutils.dumps(fake_message)]))

        self.assertEqual(fake_message, jsonutils.loads(conn.consume()[0]))
        self.assertEqual(1, len(conn.consumer.fetch_messages.mock_calls))

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, '_parse_url')
    def test_consume_timeout(self, fake_parse_url, fake_ensure_connection):
        deadline = time.time() + 3
        conn = kafka_driver.Connection(
            self.conf, '', kafka_driver.PURPOSE_LISTEN)

        conn.consumer = mock.MagicMock()
        conn.consumer.fetch_messages = mock.MagicMock(return_value=iter([]))

        self.assertRaises(driver_common.Timeout, conn.consume, timeout=3)
        self.assertEqual(0, int(deadline - time.time()))

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, '_parse_url')
    def test_consume_with_default_timeout(
            self, fake_parse_url, fake_ensure_connection):
        deadline = time.time() + 1
        conn = kafka_driver.Connection(
            self.conf, '', kafka_driver.PURPOSE_LISTEN)

        conn.consumer = mock.MagicMock()
        conn.consumer.fetch_messages = mock.MagicMock(return_value=iter([]))

        self.assertRaises(driver_common.Timeout, conn.consume)
        self.assertEqual(0, int(deadline - time.time()))

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, '_parse_url')
    def test_consume_timeout_without_consumers(
            self, fake_parse_url, fake_ensure_connection):
        deadline = time.time() + 3
        conn = kafka_driver.Connection(
            self.conf, '', kafka_driver.PURPOSE_LISTEN)
        conn.consumer = mock.MagicMock(return_value=None)

        self.assertRaises(driver_common.Timeout, conn.consume, timeout=3)
        self.assertEqual(0, int(deadline - time.time()))


class TestKafkaListener(test_utils.BaseTestCase):

    def setUp(self):
        super(TestKafkaListener, self).setUp()
        self.messaging_conf.transport_driver = 'kafka'
        transport = oslo_messaging.get_transport(self.conf)
        self.driver = transport._driver

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, 'declare_topic_consumer')
    def test_create_listener(self, fake_consumer, fake_ensure_connection):
        fake_target = oslo_messaging.Target(topic='fake_topic')
        fake_targets_and_priorities = [(fake_target, 'info')]
        self.driver.listen_for_notifications(fake_targets_and_priorities, None,
                                             None, None)
        self.assertEqual(1, len(fake_consumer.mock_calls))

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, 'declare_topic_consumer')
    def test_converting_targets_to_topics(self, fake_consumer,
                                          fake_ensure_connection):
        fake_targets_and_priorities = [
            (oslo_messaging.Target(topic="fake_topic",
                                   exchange="test1"), 'info'),
            (oslo_messaging.Target(topic="fake_topic",
                                   exchange="test2"), 'info'),
            (oslo_messaging.Target(topic="fake_topic",
                                   exchange="test1"), 'error'),
            (oslo_messaging.Target(topic="fake_topic",
                                   exchange="test3"), 'error'),
        ]
        self.driver.listen_for_notifications(fake_targets_and_priorities, None,
                                             None, None)
        self.assertEqual(1, len(fake_consumer.mock_calls))
        fake_consumer.assert_called_once_with(set(['fake_topic.error',
                                                   'fake_topic.info']),
                                              None)

    @mock.patch.object(kafka_driver.Connection, '_ensure_connection')
    @mock.patch.object(kafka_driver.Connection, 'declare_topic_consumer')
    def test_stop_listener(self, fake_consumer, fake_client):
        fake_target = oslo_messaging.Target(topic='fake_topic')
        fake_targets_and_priorities = [(fake_target, 'info')]
        listener = self.driver.listen_for_notifications(
            fake_targets_and_priorities, None, None, None)._poll_style_listener
        listener.conn.consume = mock.MagicMock()
        listener.conn.consume.return_value = (
            iter([kafka.common.KafkaMessage(
                topic='fake_topic', partition=0, offset=0,
                key=None, value='{"message": {"fake": "fake_message_1"},'
                                '"context": {"fake": "fake_context_1"}}')]))
        listener.poll()
        self.assertEqual(1, len(listener.conn.consume.mock_calls))
        listener.conn.stop_consuming = mock.MagicMock()
        listener.stop()
        fake_response = listener.poll()
        self.assertEqual(1, len(listener.conn.consume.mock_calls))
        self.assertEqual([], fake_response)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
# Copyright 2013 eNovance
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""A notification listener is used to process notification messages sent by a
notifier that uses the ``messaging`` driver.

A notification listener subscribes to the topic - and optionally exchange - in
the supplied target.  Notification messages sent by notifier clients to the
target's topic/exchange are received by the listener.

If multiple listeners subscribe to the same target, the notification will be
received by only one of the listeners. The receiving listener is selected from
the group using a best-effort round-robin algorithm.

This delivery pattern can be altered somewhat by specifying a pool name for the
listener. Listeners with the same pool name behave like a subgroup within the
group of listeners subscribed to the same topic/exchange.  Each subgroup of
listeners will receive a copy of the notification to be consumed by one member
of the subgroup. Therefore, multiple copies of the notification will be
delivered - one to the group of listeners that have no pool name (if they
exist), and one to each subgroup of listeners that share the same pool name.

Note that not all transport drivers have implemented support for listener
pools. Those drivers that do not support pools will raise a NotImplementedError
if a pool name is specified to get_notification_listener().

A notification listener exposes a number of endpoints, each of which contain a
set of methods. Each method's name corresponds to a notification's priority.
When a notification is received it is dispatched to the method named like the
notification's priority - e.g. ``info`` notifications are dispatched to the
info() method, etc.

Optionally a notification endpoint can define a NotificationFilter.
Notification messages that do not match the filter's rules will *not* be passed
to the endpoint's methods.

Parameters to endpoint methods are: the request context supplied by the client,
the publisher_id of the notification message, the event_type, the payload and
metadata. The metadata parameter is a mapping containing a unique message_id
and a timestamp.

An endpoint method can explicitly return
oslo_messaging.NotificationResult.HANDLED to acknowledge a message or
oslo_messaging.NotificationResult.REQUEUE to requeue the message. Note that not
all transport drivers implement support for requeueing. In order to use this
feature, applications should assert that the feature is available by passing
allow_requeue=True to get_notification_listener(). If the driver does not
support requeueing, it will raise NotImplementedError at this point.

The message is acknowledged only if all endpoints either return
oslo_messaging.NotificationResult.HANDLED or None.

Each notification listener is associated with an executor which controls how
incoming notification messages will be received and dispatched. By default, the
most simple executor is used - the blocking executor. This executor processes
inbound notifications on the server's thread, blocking it from processing
additional notifications until it finishes with the current one. Refer to the
Executor documentation for descriptions of the other types of executors.

*Note:* If the "eventlet" executor is used, the threading and time library need
to be monkeypatched.

Notification listener have start(), stop() and wait() messages to begin
handling requests, stop handling requests, and wait for all in-process
requests to complete after the listener has been stopped.

To create a notification listener, you supply a transport, list of targets and
a list of endpoints.

A transport can be obtained simply by calling the get_notification_transport()
method::

    transport = messaging.get_notification_transport(conf)

which will load the appropriate transport driver according to the user's
messaging configuration. See get_notification_transport() for more details.

A simple example of a notification listener with multiple endpoints might be::

    from oslo_config import cfg
    import oslo_messaging


    class NotificationEndpoint(object):
        filter_rule = oslo_messaging.NotificationFilter(
            publisher_id='^compute.*')

        def warn(self, ctxt, publisher_id, event_type, payload, metadata):
            do_something(payload)


    class ErrorEndpoint(object):
        filter_rule = oslo_messaging.NotificationFilter(
            event_type='^instance\..*\.start$',
            context={'ctxt_key': 'regexp'})

        def error(self, ctxt, publisher_id, event_type, payload, metadata):
            do_something(payload)

    transport = oslo_messaging.get_notification_transport(cfg.CONF)
    targets = [
        oslo_messaging.Target(topic='notifications'),
        oslo_messaging.Target(topic='notifications_bis')
    ]
    endpoints = [
        NotificationEndpoint(),
        ErrorEndpoint(),
    ]
    pool = "listener-workers"
    server = oslo_messaging.get_notification_listener(transport, targets,
                                                      endpoints, pool=pool)
    server.start()
    server.wait()


By supplying a serializer object, a listener can deserialize a request context
and arguments from primitive types.

"""
import itertools
import logging

from oslo_messaging._i18n import _LE
from oslo_messaging.notify import dispatcher as notify_dispatcher
from oslo_messaging import server as msg_server

LOG = logging.getLogger(__name__)


class NotificationServerBase(msg_server.MessageHandlingServer):
    def __init__(self, transport, targets, dispatcher, executor='blocking',
                 allow_requeue=True, pool=None, batch_size=1,
                 batch_timeout=None):
        super(NotificationServerBase, self).__init__(transport, dispatcher,
                                                     executor)
        self._allow_requeue = allow_requeue
        self._pool = pool
        self.targets = targets
        self._targets_priorities = set(
            itertools.product(self.targets,
                              self.dispatcher.supported_priorities)
        )

        self._batch_size = batch_size
        self._batch_timeout = batch_timeout

    def _create_listener(self):
        return self.transport._listen_for_notifications(
            self._targets_priorities, self._pool, self._batch_size,
            self._batch_timeout
        )


class NotificationServer(NotificationServerBase):
    def __init__(self, transport, targets, dispatcher, executor='blocking',
                 allow_requeue=True, pool=None):
        super(NotificationServer, self).__init__(
            transport, targets, dispatcher, executor, allow_requeue, pool, 1,
            None
        )

    def _process_incoming(self, incoming):
        message = incoming[0]
        try:
            res = self.dispatcher.dispatch(message)
        except Exception:
            LOG.exception(_LE('Exception during message handling.'))
            res = notify_dispatcher.NotificationResult.REQUEUE

        try:
            if (res == notify_dispatcher.NotificationResult.REQUEUE and
                    self._allow_requeue):
                message.requeue()
            else:
                message.acknowledge()
        except Exception:
            LOG.exception(_LE("Fail to ack/requeue message."))


class BatchNotificationServer(NotificationServerBase):

    def _process_incoming(self, incoming):
        try:
            not_processed_messages = self.dispatcher.dispatch(incoming)
        except Exception:
            not_processed_messages = set(incoming)
            LOG.exception(_LE('Exception during messages handling.'))
        for m in incoming:
            try:
                if m in not_processed_messages and self._allow_requeue:
                    m.requeue()
                else:
                    m.acknowledge()
            except Exception:
                LOG.exception(_LE("Fail to ack/requeue message."))


def get_notification_listener(transport, targets, endpoints,
                              executor='blocking', serializer=None,
                              allow_requeue=False, pool=None):
    """Construct a notification listener

    The executor parameter controls how incoming messages will be received and
    dispatched. By default, the most simple executor is used - the blocking
    executor.

    If the eventlet executor is used, the threading and time library need to be
    monkeypatched.

    :param transport: the messaging transport
    :type transport: Transport
    :param targets: the exchanges and topics to listen on
    :type targets: list of Target
    :param endpoints: a list of endpoint objects
    :type endpoints: list
    :param executor: name of a message executor - for example
                     'eventlet', 'blocking'
    :type executor: str
    :param serializer: an optional entity serializer
    :type serializer: Serializer
    :param allow_requeue: whether NotificationResult.REQUEUE support is needed
    :type allow_requeue: bool
    :param pool: the pool name
    :type pool: str
    :raises: NotImplementedError
    """
    dispatcher = notify_dispatcher.NotificationDispatcher(endpoints,
                                                          serializer)
    return NotificationServer(transport, targets, dispatcher, executor,
                              allow_requeue, pool)


def get_batch_notification_listener(transport, targets, endpoints,
                                    executor='blocking', serializer=None,
                                    allow_requeue=False, pool=None,
                                    batch_size=None, batch_timeout=None):
    """Construct a batch notification listener

    The executor parameter controls how incoming messages will be received and
    dispatched. By default, the most simple executor is used - the blocking
    executor.

    If the eventlet executor is used, the threading and time library need to be
    monkeypatched.

    :param transport: the messaging transport
    :type transport: Transport
    :param targets: the exchanges and topics to listen on
    :type targets: list of Target
    :param endpoints: a list of endpoint objects
    :type endpoints: list
    :param executor: name of a message executor - for example
                     'eventlet', 'blocking'
    :type executor: str
    :param serializer: an optional entity serializer
    :type serializer: Serializer
    :param allow_requeue: whether NotificationResult.REQUEUE support is needed
    :type allow_requeue: bool
    :param pool: the pool name
    :type pool: str
    :param batch_size: number of messages to wait before calling
                       endpoints callacks
    :type batch_size: int
    :param batch_timeout: number of seconds to wait before calling
                       endpoints callacks
    :type batch_timeout: int
    :raises: NotImplementedError
    """
    dispatcher = notify_dispatcher.BatchNotificationDispatcher(
        endpoints, serializer)
    return BatchNotificationServer(
        transport, targets, dispatcher, executor, allow_requeue, pool,
        batch_size, batch_timeout
    )



#    Copyright 2011 Cloudscaling Group, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import threading

from stevedore import driver

from oslo_messaging._drivers import base
from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client import zmq_client
from oslo_messaging._drivers.zmq_driver.server import zmq_server
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_options
from oslo_messaging._i18n import _LE


RPCException = rpc_common.RPCException


class LazyDriverItem(object):

    def __init__(self, item_cls, *args, **kwargs):
        self._lock = threading.Lock()
        self.item = None
        self.item_class = item_cls
        self.args = args
        self.kwargs = kwargs
        self.process_id = os.getpid()

    def get(self):
        #  NOTE(ozamiatin): Lazy initialization.
        #  All init stuff moved closer to usage point - lazy init.
        #  Better design approach is to initialize in the driver's
        # __init__, but 'fork' extensively used by services
        #  breaks all things.

        if self.item is not None and os.getpid() == self.process_id:
            return self.item

        with self._lock:
            if self.item is None or os.getpid() != self.process_id:
                self.process_id = os.getpid()
                self.item = self.item_class(*self.args, **self.kwargs)
        return self.item

    def cleanup(self):
        if self.item:
            self.item.cleanup()


class ZmqDriver(base.BaseDriver):

    """ZeroMQ Driver implementation.

    Provides implementation of RPC and Notifier APIs by means
    of ZeroMQ library.

    See :doc:`zmq_driver` for details.
    """

    def __init__(self, conf, url, default_exchange=None,
                 allowed_remote_exmods=None):
        """Construct ZeroMQ driver.

        Initialize driver options.

        Construct matchmaker - pluggable interface to targets management
        Name Service

        Construct client and server controllers

        :param conf: oslo messaging configuration object
        :type conf: oslo_config.CONF
        :param url: transport URL
        :type url: TransportUrl
        :param default_exchange: Not used in zmq implementation
        :type default_exchange: None
        :param allowed_remote_exmods: remote exception passing options
        :type allowed_remote_exmods: list
        """
        zmq = zmq_async.import_zmq()
        if zmq is None:
            raise ImportError(_LE("ZeroMQ is not available!"))

        zmq_options.register_opts(conf)
        self.conf = conf
        self.allowed_remote_exmods = allowed_remote_exmods

        self.matchmaker = driver.DriverManager(
            'oslo.messaging.zmq.matchmaker',
            self.get_matchmaker_backend(self.conf, url),
        ).driver(self.conf, url=url)

        client_cls = zmq_client.ZmqClientProxy
        if conf.oslo_messaging_zmq.use_pub_sub and not \
                conf.oslo_messaging_zmq.use_router_proxy:
            client_cls = zmq_client.ZmqClientMixDirectPubSub
        elif not conf.oslo_messaging_zmq.use_pub_sub and not \
                conf.oslo_messaging_zmq.use_router_proxy:
            client_cls = zmq_client.ZmqClientDirect

        self.client = LazyDriverItem(
            client_cls, self.conf, self.matchmaker,
            self.allowed_remote_exmods)

        self.notifier = LazyDriverItem(
            client_cls, self.conf, self.matchmaker,
            self.allowed_remote_exmods)

        super(ZmqDriver, self).__init__(conf, url, default_exchange,
                                        allowed_remote_exmods)

    @staticmethod
    def get_matchmaker_backend(conf, url):
        zmq_transport, _, matchmaker_backend = url.transport.partition('+')
        assert zmq_transport == 'zmq', "Needs to be zmq for this transport!"
        if not matchmaker_backend:
            return conf.oslo_messaging_zmq.rpc_zmq_matchmaker
        if matchmaker_backend not in zmq_options.MATCHMAKER_BACKENDS:
            raise rpc_common.RPCException(
                _LE("Incorrect matchmaker backend name %(backend_name)s! "
                    "Available names are: %(available_names)s") %
                {"backend_name": matchmaker_backend,
                 "available_names": zmq_options.MATCHMAKER_BACKENDS})
        return matchmaker_backend

    def send(self, target, ctxt, message, wait_for_reply=None, timeout=None,
             retry=None):
        """Send RPC message to server

        :param target: Message destination target
        :type target: oslo_messaging.Target
        :param ctxt: Message context
        :type ctxt: dict
        :param message: Message payload to pass
        :type message: dict
        :param wait_for_reply: Waiting for reply flag
        :type wait_for_reply: bool
        :param timeout: Reply waiting timeout in seconds
        :type timeout: int
        :param retry: an optional default connection retries configuration
                      None or -1 means to retry forever
                      0 means no retry
                      N means N retries
        :type retry: int
        """
        client = self.client.get()
        if wait_for_reply:
            return client.send_call(target, ctxt, message, timeout, retry)
        elif target.fanout:
            client.send_fanout(target, ctxt, message, retry)
        else:
            client.send_cast(target, ctxt, message, retry)

    def send_notification(self, target, ctxt, message, version, retry=None):
        """Send notification to server

        :param target: Message destination target
        :type target: oslo_messaging.Target
        :param ctxt: Message context
        :type ctxt: dict
        :param message: Message payload to pass
        :type message: dict
        :param version: Messaging API version
        :type version: str
        :param retry: an optional default connection retries configuration
                      None or -1 means to retry forever
                      0 means no retry
                      N means N retries
        :type retry: int
        """
        client = self.notifier.get()
        client.send_notify(target, ctxt, message, version, retry)

    def listen(self, target, batch_size, batch_timeout):
        """Listen to a specified target on a server side

        :param target: Message destination target
        :type target: oslo_messaging.Target
        """
        listener = zmq_server.ZmqServer(self, self.conf, self.matchmaker,
                                        target)
        return base.PollStyleListenerAdapter(listener, batch_size,
                                             batch_timeout)

    def listen_for_notifications(self, targets_and_priorities, pool,
                                 batch_size, batch_timeout):
        """Listen to a specified list of targets on a server side

        :param targets_and_priorities: List of pairs (target, priority)
        :type targets_and_priorities: list
        :param pool: Not used for zmq implementation
        :type pool: object
        """
        listener = zmq_server.ZmqNotificationServer(
            self, self.conf, self.matchmaker, targets_and_priorities)
        return base.PollStyleListenerAdapter(listener, batch_size,
                                             batch_timeout)

    def cleanup(self):
        """Cleanup all driver's connections finally
        """
        self.client.cleanup()
        self.notifier.cleanup()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_config import cfg

import oslo_messaging
from oslo_messaging.tests.functional import utils


class TestWithRealKafkaBroker(utils.SkipIfNoTransportURL):
    def setUp(self):
        super(TestWithRealKafkaBroker, self).setUp(conf=cfg.ConfigOpts())
        if not self.url.startswith('kafka://'):
            self.skipTest("TRANSPORT_URL is not set to kafka driver")
        transport = oslo_messaging.get_transport(self.conf, self.url)
        self.driver = transport._driver

    def test_send_and_receive_message(self):
        target = oslo_messaging.Target(
            topic="fake_topic", exchange='fake_exchange')
        targets_and_priorities = [(target, 'fake_info')]

        listener = self.driver.listen_for_notifications(
            targets_and_priorities, None, None, None)._poll_style_listener
        fake_context = {"fake_context_key": "fake_context_value"}
        fake_message = {"fake_message_key": "fake_message_value"}
        self.driver.send_notification(
            target, fake_context, fake_message, None)

        received_message = listener.poll()[0]
        self.assertEqual(fake_context, received_message.ctxt)
        self.assertEqual(fake_message, received_message.message)

    def test_send_and_receive_message_without_exchange(self):
        target = oslo_messaging.Target(topic="fake_no_exchange_topic")
        targets_and_priorities = [(target, 'fake_info')]

        listener = self.driver.listen_for_notifications(
            targets_and_priorities, None, None, None)._poll_style_listener
        fake_context = {"fake_context_key": "fake_context_value"}
        fake_message = {"fake_message_key": "fake_message_value"}
        self.driver.send_notification(
            target, fake_context, fake_message, None)

        received_message = listener.poll()[0]
        self.assertEqual(fake_context, received_message.ctxt)
        self.assertEqual(fake_message, received_message.message)

    def test_receive_message_from_empty_topic_with_timeout(self):
        target = oslo_messaging.Target(
            topic="fake_empty_topic", exchange='fake_empty_exchange')
        targets_and_priorities = [(target, 'fake_info')]

        listener = self.driver.listen_for_notifications(
            targets_and_priorities, None, None, None)._poll_style_listener

        deadline = time.time() + 3
        received_message = listener.poll(batch_timeout=3)
        self.assertEqual(0, int(deadline - time.time()))
        self.assertEqual([], received_message)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools
import time

import oslo_messaging
from oslo_messaging._drivers.zmq_driver.client import zmq_receivers
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.proxy import zmq_proxy
from oslo_messaging._drivers.zmq_driver.server.consumers.zmq_dealer_consumer \
    import DealerConsumerWithAcks
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_options
from oslo_messaging.tests.drivers.zmq import zmq_common
from oslo_messaging.tests import utils as test_utils

zmq = zmq_async.import_zmq()


class TestZmqAckManager(test_utils.BaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestZmqAckManager, self).setUp()

        # register and set necessary config opts
        self.messaging_conf.transport_driver = 'zmq'
        zmq_options.register_opts(self.conf)
        kwargs = {'rpc_zmq_matchmaker': 'dummy',
                  'use_pub_sub': False,
                  'use_router_proxy': True,
                  'rpc_thread_pool_size': 1,
                  'rpc_use_acks': True,
                  'rpc_ack_timeout_base': 5,
                  'rpc_ack_timeout_multiplier': 1,
                  'rpc_retry_attempts': 2}
        self.config(group='oslo_messaging_zmq', **kwargs)
        self.conf.register_opts(zmq_proxy.zmq_proxy_opts,
                                group='zmq_proxy_opts')

        # mock set_result method of futures
        self.set_result_patcher = mock.patch.object(
            zmq_receivers.futurist.Future, 'set_result',
            side_effect=zmq_receivers.futurist.Future.set_result, autospec=True
        )
        self.set_result = self.set_result_patcher.start()

        # mock send method of senders
        self.send_patcher = mock.patch.object(
            zmq_senders.RequestSenderProxy, 'send',
            side_effect=zmq_senders.RequestSenderProxy.send, autospec=True
        )
        self.send = self.send_patcher.start()

        # get driver
        transport = oslo_messaging.get_transport(self.conf)
        self.driver = transport._driver

        # prepare and launch proxy
        self.proxy = zmq_proxy.ZmqProxy(self.conf)
        vars(self.driver.matchmaker).update(vars(self.proxy.matchmaker))
        self.executor = zmq_async.get_executor(self.proxy.run)
        self.executor.execute()

        # create listener
        self.listener = zmq_common.TestServerListener(self.driver)

        # create target and message
        self.target = oslo_messaging.Target(topic='topic', server='server')
        self.message = {'method': 'xyz', 'args': {'x': 1, 'y': 2, 'z': 3}}

        # start listening to target
        self.listener.listen(self.target)

        # get ack manager
        self.ack_manager = self.driver.client.get().publishers['default']

        self.addCleanup(
            zmq_common.StopRpc(
                self, [('listener', 'stop'), ('executor', 'stop'),
                       ('proxy', 'close'), ('driver', 'cleanup'),
                       ('send_patcher', 'stop'),
                       ('set_result_patcher', 'stop')]
            )
        )

        # wait for all connections to be established
        # and all parties to be ready for messaging
        time.sleep(1)

    @mock.patch.object(DealerConsumerWithAcks, '_acknowledge',
                       side_effect=DealerConsumerWithAcks._acknowledge,
                       autospec=True)
    def test_cast_success_without_retries(self, received_ack_mock):
        result = self.driver.send(
            self.target, {}, self.message, wait_for_reply=False
        )
        self.assertIsNone(result)
        self.ack_manager.pool.shutdown(wait=True)
        self.assertTrue(self.listener._received.isSet())
        self.assertEqual(self.message, self.listener.message.message)
        self.assertEqual(1, self.send.call_count)
        self.assertEqual(1, received_ack_mock.call_count)
        self.assertEqual(2, self.set_result.call_count)

    def test_cast_success_with_one_retry(self):
        with mock.patch.object(DealerConsumerWithAcks,
                               '_acknowledge') as lost_ack_mock:
            result = self.driver.send(
                self.target, {}, self.message, wait_for_reply=False
            )
            self.assertIsNone(result)
            self.listener._received.wait(5)
            self.assertTrue(self.listener._received.isSet())
            self.assertEqual(self.message, self.listener.message.message)
            self.assertEqual(1, self.send.call_count)
            self.assertEqual(1, lost_ack_mock.call_count)
            self.assertEqual(0, self.set_result.call_count)
            self.listener._received.clear()
        with mock.patch.object(DealerConsumerWithAcks, '_acknowledge',
                               side_effect=DealerConsumerWithAcks._acknowledge,
                               autospec=True) as received_ack_mock:
            self.ack_manager.pool.shutdown(wait=True)
            self.assertFalse(self.listener._received.isSet())
            self.assertEqual(2, self.send.call_count)
            self.assertEqual(1, received_ack_mock.call_count)
            self.assertEqual(2, self.set_result.call_count)

    def test_cast_success_with_two_retries(self):
        with mock.patch.object(DealerConsumerWithAcks,
                               '_acknowledge') as lost_ack_mock:
            result = self.driver.send(
                self.target, {}, self.message, wait_for_reply=False
            )
            self.assertIsNone(result)
            self.listener._received.wait(5)
            self.assertTrue(self.listener._received.isSet())
            self.assertEqual(self.message, self.listener.message.message)
            self.assertEqual(1, self.send.call_count)
            self.assertEqual(1, lost_ack_mock.call_count)
            self.assertEqual(0, self.set_result.call_count)
            self.listener._received.clear()
            self.listener._received.wait(7.5)
            self.assertFalse(self.listener._received.isSet())
            self.assertEqual(2, self.send.call_count)
            self.assertEqual(2, lost_ack_mock.call_count)
            self.assertEqual(0, self.set_result.call_count)
        with mock.patch.object(DealerConsumerWithAcks, '_acknowledge',
                               side_effect=DealerConsumerWithAcks._acknowledge,
                               autospec=True) as received_ack_mock:
            self.ack_manager.pool.shutdown(wait=True)
            self.assertFalse(self.listener._received.isSet())
            self.assertEqual(3, self.send.call_count)
            self.assertEqual(1, received_ack_mock.call_count)
            self.assertEqual(2, self.set_result.call_count)

    @mock.patch.object(DealerConsumerWithAcks, '_acknowledge')
    def test_cast_failure_exhausted_retries(self, lost_ack_mock):
        result = self.driver.send(
            self.target, {}, self.message, wait_for_reply=False
        )
        self.assertIsNone(result)
        self.ack_manager.pool.shutdown(wait=True)
        self.assertTrue(self.listener._received.isSet())
        self.assertEqual(self.message, self.listener.message.message)
        self.assertEqual(3, self.send.call_count)
        self.assertEqual(3, lost_ack_mock.call_count)
        self.assertEqual(1, self.set_result.call_count)

    @mock.patch.object(DealerConsumerWithAcks, '_acknowledge',
                       side_effect=DealerConsumerWithAcks._acknowledge,
                       autospec=True)
    @mock.patch.object(DealerConsumerWithAcks, '_reply',
                       side_effect=DealerConsumerWithAcks._reply,
                       autospec=True)
    @mock.patch.object(DealerConsumerWithAcks, '_reply_from_cache',
                       side_effect=DealerConsumerWithAcks._reply_from_cache,
                       autospec=True)
    def test_call_success_without_retries(self, unused_reply_from_cache_mock,
                                          received_reply_mock,
                                          received_ack_mock):
        result = self.driver.send(
            self.target, {}, self.message, wait_for_reply=True, timeout=10
        )
        self.assertIsNotNone(result)
        self.ack_manager.pool.shutdown(wait=True)
        self.assertTrue(self.listener._received.isSet())
        self.assertEqual(self.message, self.listener.message.message)
        self.assertEqual(1, self.send.call_count)
        self.assertEqual(1, received_ack_mock.call_count)
        self.assertEqual(3, self.set_result.call_count)
        received_reply_mock.assert_called_once_with(mock.ANY, mock.ANY,
                                                    reply=True, failure=None)
        self.assertEqual(0, unused_reply_from_cache_mock.call_count)

    @mock.patch.object(DealerConsumerWithAcks, '_acknowledge')
    @mock.patch.object(DealerConsumerWithAcks, '_reply')
    @mock.patch.object(DealerConsumerWithAcks, '_reply_from_cache')
    def test_call_failure_exhausted_retries(self, lost_reply_from_cache_mock,
                                            lost_reply_mock, lost_ack_mock):
        self.assertRaises(oslo_messaging.MessagingTimeout,
                          self.driver.send,
                          self.target, {}, self.message,
                          wait_for_reply=True, timeout=20)
        self.ack_manager.pool.shutdown(wait=True)
        self.assertTrue(self.listener._received.isSet())
        self.assertEqual(self.message, self.listener.message.message)
        self.assertEqual(3, self.send.call_count)
        self.assertEqual(3, lost_ack_mock.call_count)
        self.assertEqual(2, self.set_result.call_count)
        lost_reply_mock.assert_called_once_with(mock.ANY,
                                                reply=True, failure=None)
        self.assertEqual(2, lost_reply_from_cache_mock.call_count)



# Copyright 2015 Mirantis Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_context import context as common_context
from six.moves import mock

from oslo_messaging import serializer
from oslo_messaging.tests import utils as test_utils


class TestRequestContextSerializer(test_utils.BaseTestCase):

    def setUp(self):
        super(TestRequestContextSerializer, self).setUp()

        self.serializer = serializer.RequestContextSerializer(mock.MagicMock())
        self.context = common_context.RequestContext()
        self.entity = {'foo': 'bar'}

    def test_serialize_entity(self):
        self.serializer.serialize_entity(self.context, self.entity)
        self.serializer._base.serialize_entity.assert_called_with(
            self.context, self.entity)

    def test_serialize_entity_empty_base(self):
        # NOTE(viktors): Return False for check `if self.serializer._base:`
        bool_args = {'__bool__': lambda *args: False,
                     '__nonzero__': lambda *args: False}
        self.serializer._base.configure_mock(**bool_args)

        entity = self.serializer.serialize_entity(self.context, self.entity)
        self.assertFalse(self.serializer._base.serialize_entity.called)
        self.assertEqual(self.entity, entity)

    def test_deserialize_entity(self):
        self.serializer.deserialize_entity(self.context, self.entity)
        self.serializer._base.deserialize_entity.assert_called_with(
            self.context, self.entity)

    def test_deserialize_entity_empty_base(self):
        # NOTE(viktors): Return False for check `if self.serializer._base:`
        bool_args = {'__bool__': lambda *args: False,
                     '__nonzero__': lambda *args: False}
        self.serializer._base.configure_mock(**bool_args)

        entity = self.serializer.deserialize_entity(self.context, self.entity)
        self.assertFalse(self.serializer._base.serialize_entity.called)
        self.assertEqual(self.entity, entity)

    def test_serialize_context(self):
        new_context = self.serializer.serialize_context(self.context)

        self.assertEqual(self.context.to_dict(), new_context)

    @mock.patch.object(common_context.RequestContext, 'from_dict',
                       return_value='foobar')
    def test_deserialize_context(self, mock_to_dict):
        new_context = self.serializer.deserialize_context(self.context)

        mock_to_dict.assert_called_with(self.context)
        self.assertEqual(
            common_context.RequestContext.from_dict(self.context),
            new_context
        )



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging._drivers.zmq_driver.client.publishers.dealer \
    import zmq_dealer_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_receivers
from oslo_messaging._drivers.zmq_driver.client import zmq_routing_table
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

zmq = zmq_async.import_zmq()


class DealerPublisherDirect(zmq_dealer_publisher_base.DealerPublisherBase):
    """DEALER-publisher using direct connections.

    Publishing directly to remote services assumes the following:
        -   All direct connections are dynamic - so they live per message,
            thus each message send executes the following:
                *   Open a new socket
                *   Connect to some host got from the RoutingTable
                *   Send message(s)
                *   Close connection, destroy socket
        -   RoutingTable/RoutingTableUpdater implements local cache of
            matchmaker (e.g. Redis) for target resolution to the list of
            available hosts. Cache updates in a background thread.
        -   Caching of connections is not appropriate for directly connected
            OS services, because finally it results in a full-mesh of
            connections between services.
        -   Yes we lose on performance opening and closing connections
            for each message, but that is done intentionally to implement
            the dynamic connections concept. The key thought here is to
            have minimum number of connected services at the moment.
        -   Using the local RoutingTable cache is done to optimise access
            to the matchmaker so we don't call the matchmaker per each message
    """

    def __init__(self, conf, matchmaker):
        sender = zmq_senders.RequestSenderDirect(conf)
        receiver = zmq_receivers.ReceiverDirect(conf)
        super(DealerPublisherDirect, self).__init__(conf, matchmaker,
                                                    sender, receiver)

        self.routing_table = zmq_routing_table.RoutingTableAdaptor(
            conf, matchmaker, zmq.ROUTER)

    def _get_round_robin_host_connection(self, target, socket):
        host = self.routing_table.get_round_robin_host(target)
        socket.connect_to_host(host)

    def _get_fanout_connection(self, target, socket):
        for host in self.routing_table.get_fanout_hosts(target):
            socket.connect_to_host(host)

    def acquire_connection(self, request):
        socket = self.sockets_manager.get_socket()
        if request.msg_type in zmq_names.DIRECT_TYPES:
            self._get_round_robin_host_connection(request.target, socket)
        elif request.msg_type in zmq_names.MULTISEND_TYPES:
            self._get_fanout_connection(request.target, socket)
        return socket

    def _finally_unregister(self, socket, request):
        super(DealerPublisherDirect, self)._finally_unregister(socket, request)
        self.receiver.unregister_socket(socket)

    def send_request(self, socket, request):
        if request.msg_type in zmq_names.MULTISEND_TYPES:
            for _ in range(socket.connections_count()):
                self.sender.send(socket, request)
        else:
            self.sender.send(socket, request)

    def cleanup(self):
        self.routing_table.cleanup()
        super(DealerPublisherDirect, self).cleanup()



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import uuid

import six

from oslo_messaging._drivers.zmq_driver.client.publishers.dealer \
    import zmq_dealer_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_receivers
from oslo_messaging._drivers.zmq_driver.client import zmq_routing_table
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.matchmaker import zmq_matchmaker_base
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_updater

zmq = zmq_async.import_zmq()


class DealerPublisherProxy(zmq_dealer_publisher_base.DealerPublisherBase):
    """DEALER-publisher via proxy."""

    def __init__(self, conf, matchmaker):
        sender = zmq_senders.RequestSenderProxy(conf)
        receiver = zmq_receivers.ReceiverProxy(conf)
        super(DealerPublisherProxy, self).__init__(conf, matchmaker,
                                                   sender, receiver)

        self.socket = self.sockets_manager.get_socket_to_publishers(
            self._generate_identity())
        self.routing_table = zmq_routing_table.RoutingTableAdaptor(
            conf, matchmaker, zmq.DEALER)
        self.connection_updater = PublisherConnectionUpdater(
            self.conf, self.matchmaker, self.socket)

    def _generate_identity(self):
        return six.b(self.conf.oslo_messaging_zmq.rpc_zmq_host + "/" +
                     str(uuid.uuid4()))

    def _check_reply(self, reply, request):
        super(DealerPublisherProxy, self)._check_reply(reply, request)
        assert reply.reply_id == request.routing_key, \
            "Reply from recipient expected!"

    def _get_routing_keys(self, request):
        if request.msg_type in zmq_names.DIRECT_TYPES:
            return [self.routing_table.get_round_robin_host(request.target)]
        else:
            return \
                [zmq_address.target_to_subscribe_filter(request.target)] \
                if self.conf.oslo_messaging_zmq.use_pub_sub else \
                self.routing_table.get_fanout_hosts(request.target)

    def acquire_connection(self, request):
        return self.socket

    def send_request(self, socket, request):
        for routing_key in self._get_routing_keys(request):
            request.routing_key = routing_key
            self.sender.send(socket, request)

    def cleanup(self):
        self.connection_updater.stop()
        self.routing_table.cleanup()
        self.socket.close()
        super(DealerPublisherProxy, self).cleanup()


class PublisherConnectionUpdater(zmq_updater.ConnectionUpdater):

    def _update_connection(self):
        publishers = self.matchmaker.get_publishers()
        for pub_address, fe_router_address in publishers:
            self.socket.connect_to_host(fe_router_address)


class DealerPublisherProxyDynamic(
        zmq_dealer_publisher_base.DealerPublisherBase):

    def __init__(self, conf, matchmaker):
        sender = zmq_senders.RequestSenderProxy(conf)
        receiver = zmq_receivers.ReceiverDirect(conf)
        super(DealerPublisherProxyDynamic, self).__init__(conf, matchmaker,
                                                          sender, receiver)

        self.publishers = set()
        self.updater = DynamicPublishersUpdater(conf, matchmaker,
                                                self.publishers)
        self.updater.update_publishers()

    def acquire_connection(self, request):
        if not self.publishers:
            raise zmq_matchmaker_base.MatchmakerUnavailable()
        socket = self.sockets_manager.get_socket()
        socket.connect_to_host(random.choice(tuple(self.publishers)))
        return socket

    def send_request(self, socket, request):
        request.routing_key = \
            zmq_address.target_to_subscribe_filter(request.target)
        self.sender.send(socket, request)

    def cleanup(self):
        self.updater.cleanup()
        super(DealerPublisherProxyDynamic, self).cleanup()


class DynamicPublishersUpdater(zmq_updater.UpdaterBase):

    def __init__(self, conf, matchmaker, publishers):
        super(DynamicPublishersUpdater, self).__init__(
            conf, matchmaker, self.update_publishers,
            sleep_for=conf.oslo_messaging_zmq.zmq_target_update
        )
        self.publishers = publishers

    def update_publishers(self):
        publishers = self.matchmaker.get_publishers()
        for pub_address, fe_router_address in publishers:
            self.publishers.add(fe_router_address)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_socket

zmq = zmq_async.import_zmq()


class SocketsManager(object):

    def __init__(self, conf, matchmaker, socket_type):
        self.conf = conf
        self.matchmaker = matchmaker
        self.socket_type = socket_type
        self.zmq_context = zmq.Context()
        self.socket_to_publishers = None
        self.socket_to_routers = None

    def get_socket(self):
        socket = zmq_socket.ZmqSocket(self.conf, self.zmq_context,
                                      self.socket_type, immediate=False)
        return socket

    def get_socket_to_publishers(self, identity=None):
        if self.socket_to_publishers is not None:
            return self.socket_to_publishers
        self.socket_to_publishers = zmq_socket.ZmqSocket(
            self.conf, self.zmq_context, self.socket_type,
            immediate=self.conf.oslo_messaging_zmq.zmq_immediate,
            identity=identity)
        publishers = self.matchmaker.get_publishers()
        for pub_address, fe_router_address in publishers:
            self.socket_to_publishers.connect_to_host(fe_router_address)
        return self.socket_to_publishers

    def get_socket_to_routers(self, identity=None):
        if self.socket_to_routers is not None:
            return self.socket_to_routers
        self.socket_to_routers = zmq_socket.ZmqSocket(
            self.conf, self.zmq_context, self.socket_type,
            immediate=self.conf.oslo_messaging_zmq.zmq_immediate,
            identity=identity)
        routers = self.matchmaker.get_routers()
        for be_router_address in routers:
            self.socket_to_routers.connect_to_host(be_router_address)
        return self.socket_to_routers



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from oslo_config import cfg

from oslo_messaging._drivers import base
from oslo_messaging import server


MATCHMAKER_BACKENDS = ('redis', 'sentinel', 'dummy')
MATCHMAKER_DEFAULT = 'redis'


zmq_opts = [
    cfg.StrOpt('rpc_zmq_bind_address', default='*',
               deprecated_group='DEFAULT',
               help='ZeroMQ bind address. Should be a wildcard (*), '
                    'an ethernet interface, or IP. '
                    'The "host" option should point or resolve to this '
                    'address.'),

    cfg.StrOpt('rpc_zmq_matchmaker', default=MATCHMAKER_DEFAULT,
               choices=MATCHMAKER_BACKENDS,
               deprecated_group='DEFAULT',
               help='MatchMaker driver.'),

    cfg.IntOpt('rpc_zmq_contexts', default=1,
               deprecated_group='DEFAULT',
               help='Number of ZeroMQ contexts, defaults to 1.'),

    cfg.IntOpt('rpc_zmq_topic_backlog',
               deprecated_group='DEFAULT',
               help='Maximum number of ingress messages to locally buffer '
                    'per topic. Default is unlimited.'),

    cfg.StrOpt('rpc_zmq_ipc_dir', default='/var/run/openstack',
               deprecated_group='DEFAULT',
               help='Directory for holding IPC sockets.'),

    cfg.StrOpt('rpc_zmq_host', default=socket.gethostname(),
               sample_default='localhost',
               deprecated_group='DEFAULT',
               help='Name of this node. Must be a valid hostname, FQDN, or '
                    'IP address. Must match "host" option, if running Nova.'),

    cfg.IntOpt('zmq_linger', default=-1,
               deprecated_group='DEFAULT',
               deprecated_name='rpc_cast_timeout',
               help='Number of seconds to wait before all pending '
                    'messages will be sent after closing a socket. '
                    'The default value of -1 specifies an infinite linger '
                    'period. The value of 0 specifies no linger period. '
                    'Pending messages shall be discarded immediately '
                    'when the socket is closed. Positive values specify an '
                    'upper bound for the linger period.'),

    cfg.IntOpt('rpc_poll_timeout', default=1,
               deprecated_group='DEFAULT',
               help='The default number of seconds that poll should wait. '
                    'Poll raises timeout exception when timeout expired.'),

    cfg.IntOpt('zmq_target_expire', default=300,
               deprecated_group='DEFAULT',
               help='Expiration timeout in seconds of a name service record '
                    'about existing target ( < 0 means no timeout).'),

    cfg.IntOpt('zmq_target_update', default=180,
               deprecated_group='DEFAULT',
               help='Update period in seconds of a name service record '
                    'about existing target.'),

    cfg.BoolOpt('use_pub_sub', default=False,
                deprecated_group='DEFAULT',
                help='Use PUB/SUB pattern for fanout methods. '
                     'PUB/SUB always uses proxy.'),

    cfg.BoolOpt('use_router_proxy', default=True,
                deprecated_group='DEFAULT',
                help='Use ROUTER remote proxy.'),

    cfg.PortOpt('rpc_zmq_min_port',
                default=49153,
                deprecated_group='DEFAULT',
                help='Minimal port number for random ports range.'),

    cfg.IntOpt('rpc_zmq_max_port',
               min=1,
               max=65536,
               default=65536,
               deprecated_group='DEFAULT',
               help='Maximal port number for random ports range.'),

    cfg.IntOpt('rpc_zmq_bind_port_retries',
               default=100,
               deprecated_group='DEFAULT',
               help='Number of retries to find free port number before '
                    'fail with ZMQBindError.'),

    cfg.StrOpt('rpc_zmq_serialization', default='json',
               choices=('json', 'msgpack'),
               deprecated_group='DEFAULT',
               help='Default serialization mechanism for '
                    'serializing/deserializing outgoing/incoming messages'),

    cfg.BoolOpt('zmq_immediate', default=True,
                help='This option configures round-robin mode in zmq socket. '
                     'True means not keeping a queue when server side '
                     'disconnects. False means to keep queue and messages '
                     'even if server is disconnected, when the server '
                     'appears we send all accumulated messages to it.'),

    cfg.IntOpt('zmq_tcp_keepalive', default=-1,
               help='Enable/disable TCP keepalive (KA) mechanism. '
                    'The default value of -1 (or any other negative value) '
                    'means to skip any overrides and leave it to OS default; '
                    '0 and 1 (or any other positive value) mean to '
                    'disable and enable the option respectively.'),

    cfg.IntOpt('zmq_tcp_keepalive_idle', default=-1,
               help='The duration between two keepalive transmissions in '
                    'idle condition. '
                    'The unit is platform dependent, for example, '
                    'seconds in Linux, milliseconds in Windows etc. '
                    'The default value of -1 (or any other negative value '
                    'and 0) means to skip any overrides and leave it '
                    'to OS default.'),

    cfg.IntOpt('zmq_tcp_keepalive_cnt', default=-1,
               help='The number of retransmissions to be carried out before '
                    'declaring that remote end is not available. '
                    'The default value of -1 (or any other negative value '
                    'and 0) means to skip any overrides and leave it '
                    'to OS default.'),

    cfg.IntOpt('zmq_tcp_keepalive_intvl', default=-1,
               help='The duration between two successive keepalive '
                    'retransmissions, if acknowledgement to the previous '
                    'keepalive transmission is not received. '
                    'The unit is platform dependent, for example, '
                    'seconds in Linux, milliseconds in Windows etc. '
                    'The default value of -1 (or any other negative value '
                    'and 0) means to skip any overrides and leave it '
                    'to OS default.'),

    cfg.IntOpt('rpc_thread_pool_size', default=100,
               help='Maximum number of (green) threads to work concurrently.'),

    cfg.IntOpt('rpc_message_ttl', default=300,
               help='Expiration timeout in seconds of a sent/received message '
                    'after which it is not tracked anymore by a '
                    'client/server.'),

    cfg.BoolOpt('rpc_use_acks', default=False,
                help='Wait for message acknowledgements from receivers. '
                     'This mechanism works only via proxy without PUB/SUB.'),

    cfg.IntOpt('rpc_ack_timeout_base', default=15,
               help='Number of seconds to wait for an ack from a cast/call. '
                    'After each retry attempt this timeout is multiplied by '
                    'some specified multiplier.'),

    cfg.IntOpt('rpc_ack_timeout_multiplier', default=2,
               help='Number to multiply base ack timeout by after each retry '
                    'attempt.'),

    cfg.IntOpt('rpc_retry_attempts', default=3,
               help='Default number of message sending attempts in case '
                    'of any problems occurred: positive value N means '
                    'at most N retries, 0 means no retries, None or -1 '
                    '(or any other negative values) mean to retry forever. '
                    'This option is used only if acknowledgments are '
                    'enabled.'),

    cfg.ListOpt('subscribe_on',
                default=[],
                help='List of publisher hosts SubConsumer can subscribe on. '
                     'This option has higher priority then the default '
                     'publishers list taken from the matchmaker.'),
]


def register_opts(conf):
    opt_group = cfg.OptGroup(name='oslo_messaging_zmq',
                             title='ZeroMQ driver options')
    conf.register_opts(zmq_opts, group=opt_group)
    conf.register_opts(server._pool_opts)
    conf.register_opts(base.base_opts)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from concurrent import futures
import logging

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client.publishers \
    import zmq_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_response
from oslo_messaging._drivers.zmq_driver.client import zmq_sockets_manager
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._i18n import _LE

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerPublisherBase(zmq_publisher_base.PublisherBase):
    """Abstract DEALER-publisher."""

    def __init__(self, conf, matchmaker, sender, receiver):
        sockets_manager = zmq_sockets_manager.SocketsManager(
            conf, matchmaker, zmq.DEALER)
        super(DealerPublisherBase, self).__init__(
            sockets_manager, sender, receiver)

    def _check_reply(self, reply, request):
        assert isinstance(reply, zmq_response.Reply), "Reply expected!"

    def _finally_unregister(self, socket, request):
        self.receiver.untrack_request(request)

    def receive_reply(self, socket, request):
        self.receiver.register_socket(socket)
        _, reply_future = self.receiver.track_request(request)

        try:
            reply = reply_future.result(timeout=request.timeout)
            self._check_reply(reply, request)
        except AssertionError:
            LOG.error(_LE("Message format error in reply for %s"),
                      request.message_id)
            return None
        except futures.TimeoutError:
            self._raise_timeout(request)
        finally:
            self._finally_unregister(socket, request)

        if reply.failure:
            raise rpc_common.deserialize_remote_exception(
                reply.failure, request.allowed_remote_exmods)
        else:
            return reply.reply_body



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

import ast
import six

oslo_namespace_imports_dot = re.compile(r"import[\s]+oslo[.][^\s]+")
oslo_namespace_imports_from_dot = re.compile(r"from[\s]+oslo[.]")
oslo_namespace_imports_from_root = re.compile(r"from[\s]+oslo[\s]+import[\s]+")


def check_oslo_namespace_imports(logical_line):
    if re.match(oslo_namespace_imports_from_dot, logical_line):
        msg = ("O323: '%s' must be used instead of '%s'.") % (
            logical_line.replace('oslo.', 'oslo_'),
            logical_line)
        yield(0, msg)
    elif re.match(oslo_namespace_imports_from_root, logical_line):
        msg = ("O323: '%s' must be used instead of '%s'.") % (
            logical_line.replace('from oslo import ', 'import oslo_'),
            logical_line)
        yield(0, msg)
    elif re.match(oslo_namespace_imports_dot, logical_line):
        msg = ("O323: '%s' must be used instead of '%s'.") % (
            logical_line.replace('import', 'from').replace('.', ' import '),
            logical_line)
        yield(0, msg)


class BaseASTChecker(ast.NodeVisitor):
    """Provides a simple framework for writing AST-based checks.

    Subclasses should implement visit_* methods like any other AST visitor
    implementation. When they detect an error for a particular node the
    method should call ``self.add_error(offending_node)``. Details about
    where in the code the error occurred will be pulled from the node
    object.

    Subclasses should also provide a class variable named CHECK_DESC to
    be used for the human readable error message.

    """

    def __init__(self, tree, filename):
        """This object is created automatically by pep8.

        :param tree: an AST tree
        :param filename: name of the file being analyzed
                         (ignored by our checks)
        """
        self._tree = tree
        self._errors = []

    def run(self):
        """Called automatically by pep8."""
        self.visit(self._tree)
        return self._errors

    def add_error(self, node, message=None):
        """Add an error caused by a node to the list of errors for pep8."""
        message = message or self.CHECK_DESC
        error = (node.lineno, node.col_offset, message, self.__class__)
        self._errors.append(error)


class CheckForLoggingIssues(BaseASTChecker):

    DEBUG_CHECK_DESC = 'O324 Using translated string in debug logging'
    NONDEBUG_CHECK_DESC = 'O325 Not using translating helper for logging'
    EXCESS_HELPER_CHECK_DESC = 'O326 Using hints when _ is necessary'
    LOG_MODULES = ('logging')
    I18N_MODULES = (
        'oslo_messaging._i18n._',
        'oslo_messaging._i18n._LI',
        'oslo_messaging._i18n._LW',
        'oslo_messaging._i18n._LE',
        'oslo_messaging._i18n._LC',
    )
    TRANS_HELPER_MAP = {
        'debug': None,
        'info': '_LI',
        'warn': '_LW',
        'warning': '_LW',
        'error': '_LE',
        'exception': '_LE',
        'critical': '_LC',
    }

    def __init__(self, tree, filename):
        super(CheckForLoggingIssues, self).__init__(tree, filename)

        self.logger_names = []
        self.logger_module_names = []
        self.i18n_names = {}

        # NOTE(dstanek): this kinda accounts for scopes when talking
        # about only leaf node in the graph
        self.assignments = {}

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        item._parent = node
                        self.visit(item)
            elif isinstance(value, ast.AST):
                value._parent = node
                self.visit(value)

    def _filter_imports(self, module_name, alias):
        """Keeps lists of logging and i18n imports."""
        if module_name in self.LOG_MODULES:
            self.logger_module_names.append(alias.asname or alias.name)
        elif module_name in self.I18N_MODULES:
            self.i18n_names[alias.asname or alias.name] = alias.name

    def visit_Import(self, node):
        for alias in node.names:
            self._filter_imports(alias.name, alias)
        return super(CheckForLoggingIssues, self).generic_visit(node)

    def visit_ImportFrom(self, node):
        for alias in node.names:
            full_name = '%s.%s' % (node.module, alias.name)
            self._filter_imports(full_name, alias)
        return super(CheckForLoggingIssues, self).generic_visit(node)

    def _find_name(self, node):
        """Return the fully qualified name or a Name or Attribute."""
        if isinstance(node, ast.Name):
            return node.id
        elif (isinstance(node, ast.Attribute)
                and isinstance(node.value, (ast.Name, ast.Attribute))):
            method_name = node.attr
            obj_name = self._find_name(node.value)
            if obj_name is None:
                return None
            return obj_name + '.' + method_name
        elif isinstance(node, six.string_types):
            return node
        else:  # could be Subscript, Call or many more
            return None

    def visit_Assign(self, node):
        """Look for 'LOG = logging.getLogger'

        This handles the simple case:
          name = [logging_module].getLogger(...)

          - or -

          name = [i18n_name](...)

        And some much more comple ones:
          name = [i18n_name](...) % X

          - or -

          self.name = [i18n_name](...) % X

        """
        attr_node_types = (ast.Name, ast.Attribute)

        if (len(node.targets) != 1
                or not isinstance(node.targets[0], attr_node_types)):
            # say no to: "x, y = ..."
            return super(CheckForLoggingIssues, self).generic_visit(node)

        target_name = self._find_name(node.targets[0])

        if (isinstance(node.value, ast.BinOp) and
                isinstance(node.value.op, ast.Mod)):
            if (isinstance(node.value.left, ast.Call) and
                    isinstance(node.value.left.func, ast.Name) and
                    node.value.left.func.id in self.i18n_names):
                # NOTE(dstanek): this is done to match cases like:
                # `msg = _('something %s') % x`
                node = ast.Assign(value=node.value.left)

        if not isinstance(node.value, ast.Call):
            # node.value must be a call to getLogger
            self.assignments.pop(target_name, None)
            return super(CheckForLoggingIssues, self).generic_visit(node)

        # is this a call to an i18n function?
        if (isinstance(node.value.func, ast.Name)
                and node.value.func.id in self.i18n_names):
            self.assignments[target_name] = node.value.func.id
            return super(CheckForLoggingIssues, self).generic_visit(node)

        if (not isinstance(node.value.func, ast.Attribute)
                or not isinstance(node.value.func.value, attr_node_types)):
            # function must be an attribute on an object like
            # logging.getLogger
            return super(CheckForLoggingIssues, self).generic_visit(node)

        object_name = self._find_name(node.value.func.value)
        func_name = node.value.func.attr

        if (object_name in self.logger_module_names
                and func_name == 'getLogger'):
            self.logger_names.append(target_name)

        return super(CheckForLoggingIssues, self).generic_visit(node)

    def visit_Call(self, node):
        """Look for the 'LOG.*' calls."""
        # obj.method
        if isinstance(node.func, ast.Attribute):
            obj_name = self._find_name(node.func.value)
            if isinstance(node.func.value, ast.Name):
                method_name = node.func.attr
            elif isinstance(node.func.value, ast.Attribute):
                obj_name = self._find_name(node.func.value)
                method_name = node.func.attr
            else:  # could be Subscript, Call or many more
                return super(CheckForLoggingIssues, self).generic_visit(node)

            # if dealing with a logger the method can't be "warn"
            if obj_name in self.logger_names and method_name == 'warn':
                msg = node.args[0]  # first arg to a logging method is the msg
                self.add_error(msg, message=self.USING_DEPRECATED_WARN)

            # must be a logger instance and one of the support logging methods
            if (obj_name not in self.logger_names
                    or method_name not in self.TRANS_HELPER_MAP):
                return super(CheckForLoggingIssues, self).generic_visit(node)

            # the call must have arguments
            if not node.args:
                return super(CheckForLoggingIssues, self).generic_visit(node)

            if method_name == 'debug':
                self._process_debug(node)
            elif method_name in self.TRANS_HELPER_MAP:
                self._process_non_debug(node, method_name)

        return super(CheckForLoggingIssues, self).generic_visit(node)

    def _process_debug(self, node):
        msg = node.args[0]  # first arg to a logging method is the msg

        # if first arg is a call to a i18n name
        if (isinstance(msg, ast.Call)
                and isinstance(msg.func, ast.Name)
                and msg.func.id in self.i18n_names):
            self.add_error(msg, message=self.DEBUG_CHECK_DESC)

        # if the first arg is a reference to a i18n call
        elif (isinstance(msg, ast.Name)
                and msg.id in self.assignments
                and not self._is_raised_later(node, msg.id)):
            self.add_error(msg, message=self.DEBUG_CHECK_DESC)

    def _process_non_debug(self, node, method_name):
        msg = node.args[0]  # first arg to a logging method is the msg

        # if first arg is a call to a i18n name
        if isinstance(msg, ast.Call):
            try:
                func_name = msg.func.id
            except AttributeError:
                # in the case of logging only an exception, the msg function
                # will not have an id associated with it, for instance:
                # LOG.warning(six.text_type(e))
                return

            # the function name is the correct translation helper
            # for the logging method
            if func_name == self.TRANS_HELPER_MAP[method_name]:
                return

            # the function name is an alias for the correct translation
            # helper for the loggine method
            if (self.i18n_names[func_name] ==
                    self.TRANS_HELPER_MAP[method_name]):
                return

            self.add_error(msg, message=self.NONDEBUG_CHECK_DESC)

        # if the first arg is not a reference to the correct i18n hint
        elif isinstance(msg, ast.Name):

            # FIXME(dstanek): to make sure more robust we should be checking
            # all names passed into a logging method. we can't right now
            # because:
            # 1. We have code like this that we'll fix when dealing with the %:
            #       msg = _('....') % {}
            #       LOG.warning(msg)
            # 2. We also do LOG.exception(e) in several places. I'm not sure
            #    exactly what we should be doing about that.
            if msg.id not in self.assignments:
                return

            helper_method_name = self.TRANS_HELPER_MAP[method_name]
            if (self.assignments[msg.id] != helper_method_name
                    and not self._is_raised_later(node, msg.id)):
                self.add_error(msg, message=self.NONDEBUG_CHECK_DESC)
            elif (self.assignments[msg.id] == helper_method_name
                    and self._is_raised_later(node, msg.id)):
                self.add_error(msg, message=self.EXCESS_HELPER_CHECK_DESC)

    def _is_raised_later(self, node, name):

        def find_peers(node):
            node_for_line = node._parent
            for _field, value in ast.iter_fields(node._parent._parent):
                if isinstance(value, list) and node_for_line in value:
                    return value[value.index(node_for_line) + 1:]
                continue
            return []

        peers = find_peers(node)
        for peer in peers:
            if isinstance(peer, ast.Raise):
                if six.PY3:
                    exc = peer.exc
                else:
                    exc = peer.type
                if (isinstance(exc, ast.Call) and
                        len(exc.args) > 0 and
                        isinstance(exc.args[0], ast.Name) and
                        name in (a.id for a in exc.args)):
                    return True
                else:
                    return False
            elif isinstance(peer, ast.Assign):
                if name in (t.id for t in peer.targets if hasattr(t, 'id')):
                    return False


def factory(register):
    register(CheckForLoggingIssues)
    register(check_oslo_namespace_imports)



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six

import oslo_messaging
from oslo_messaging._drivers.zmq_driver import zmq_async

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class PublisherBase(object):

    """Abstract publisher class

    Each publisher from zmq-driver client should implement
    this interface to serve as a messages publisher.

    Publisher can send request objects from zmq_request.
    """

    def __init__(self, sockets_manager, sender, receiver):

        """Construct publisher.

        Accept sockets manager, sender and receiver objects.

        :param sockets_manager: sockets manager object
        :type sockets_manager: zmq_sockets_manager.SocketsManager
        :param sender: request sender object
        :type sender: zmq_senders.RequestSenderBase
        :param receiver: response receiver object
        :type receiver: zmq_receivers.ReceiverBase
        """
        self.sockets_manager = sockets_manager
        self.conf = sockets_manager.conf
        self.matchmaker = sockets_manager.matchmaker
        self.sender = sender
        self.receiver = receiver

    @abc.abstractmethod
    def acquire_connection(self, request):
        """Get socket to publish request on it.

        :param request: request object
        :type senders: zmq_request.Request
        """

    @abc.abstractmethod
    def send_request(self, socket, request):
        """Publish request on a socket.

        :param socket: socket object to publish request on
        :type socket: zmq_socket.ZmqSocket
        :param request: request object
        :type senders: zmq_request.Request
        """

    @abc.abstractmethod
    def receive_reply(self, socket, request):
        """Wait for a reply via the socket used for sending the request.

        :param socket: socket object to receive reply from
        :type socket: zmq_socket.ZmqSocket
        :param request: request object
        :type senders: zmq_request.Request
        """

    @staticmethod
    def _raise_timeout(request):
        raise oslo_messaging.MessagingTimeout(
            "Timeout %(tout)s seconds was reached for message %(msg_id)s" %
            {"tout": request.timeout, "msg_id": request.message_id}
        )

    def cleanup(self):
        """Cleanup publisher: stop receiving responses, close allocated
        connections etc.
        """
        self.receiver.stop()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import oslo_messaging
from oslo_messaging._drivers import common
from oslo_messaging._drivers.zmq_driver.matchmaker.zmq_matchmaker_base \
    import MatchmakerDummy
from oslo_messaging._drivers.zmq_driver.matchmaker import zmq_matchmaker_redis
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging.tests import utils as test_utils

zmq = zmq_async.import_zmq()

redis = zmq_matchmaker_redis.redis
sentinel = zmq_matchmaker_redis.redis_sentinel


class TestZmqTransportUrl(test_utils.BaseTestCase):

    @testtools.skipIf(zmq is None, "zmq not available")
    def setUp(self):
        super(TestZmqTransportUrl, self).setUp()

    def setup_url(self, url):
        transport = oslo_messaging.get_transport(self.conf, url)
        self.addCleanup(transport.cleanup)
        driver = transport._driver
        return driver, url

    def mock_redis(self):
        if redis is None:
            self.skipTest("redis not available")
        else:
            redis_patcher = mock.patch.object(redis, 'StrictRedis')
            self.addCleanup(redis_patcher.stop)
            return redis_patcher.start()

    def mock_sentinel(self):
        if sentinel is None:
            self.skipTest("sentinel not available")
        else:
            sentinel_patcher = mock.patch.object(sentinel, 'Sentinel')
            self.addCleanup(sentinel_patcher.stop)
            return sentinel_patcher.start()

    def test_empty_url(self):
        self.mock_redis()
        driver, url = self.setup_url("zmq:///")
        self.assertIs(zmq_matchmaker_redis.MatchmakerRedis,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq', driver.matchmaker.url.transport)

    def test_error_url(self):
        self.assertRaises(common.RPCException, self.setup_url, "zmq+error:///")

    def test_dummy_url(self):
        driver, url = self.setup_url("zmq+dummy:///")
        self.assertIs(MatchmakerDummy,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq+dummy', driver.matchmaker.url.transport)

    def test_redis_url(self):
        self.mock_redis()
        driver, url = self.setup_url("zmq+redis:///")
        self.assertIs(zmq_matchmaker_redis.MatchmakerRedis,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq+redis', driver.matchmaker.url.transport)

    def test_sentinel_url(self):
        self.mock_sentinel()
        driver, url = self.setup_url("zmq+sentinel:///")
        self.assertIs(zmq_matchmaker_redis.MatchmakerSentinel,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq+sentinel', driver.matchmaker.url.transport)

    def test_host_with_credentials_url(self):
        self.mock_redis()
        driver, url = self.setup_url("zmq://:password@host:60000/")
        self.assertIs(zmq_matchmaker_redis.MatchmakerRedis,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq', driver.matchmaker.url.transport)
        self.assertEqual(
            [{"host": "host", "port": 60000, "password": "password"}],
            driver.matchmaker._redis_hosts
        )

    def test_redis_multiple_hosts_url(self):
        self.mock_redis()
        driver, url = self.setup_url(
            "zmq+redis://host1:60001,host2:60002,host3:60003/"
        )
        self.assertIs(zmq_matchmaker_redis.MatchmakerRedis,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq+redis', driver.matchmaker.url.transport)
        self.assertEqual(
            [{"host": "host1", "port": 60001, "password": None},
             {"host": "host2", "port": 60002, "password": None},
             {"host": "host3", "port": 60003, "password": None}],
            driver.matchmaker._redis_hosts
        )

    def test_sentinel_multiple_hosts_url(self):
        self.mock_sentinel()
        driver, url = self.setup_url(
            "zmq+sentinel://host1:20001,host2:20002,host3:20003/"
        )
        self.assertIs(zmq_matchmaker_redis.MatchmakerSentinel,
                      driver.matchmaker.__class__)
        self.assertEqual('zmq+sentinel', driver.matchmaker.url.transport)
        self.assertEqual(
            [("host1", 20001), ("host2", 20002), ("host3", 20003)],
            driver.matchmaker._sentinel_hosts
        )




# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def version_is_compatible(imp_version, version):
    """Determine whether versions are compatible.

    :param imp_version: The version implemented
    :param version: The version requested by an incoming message.
    """
    if imp_version is None:
        return True

    if version is None:
        return False

    version_parts = version.split('.')
    imp_version_parts = imp_version.split('.')
    try:
        rev = version_parts[2]
    except IndexError:
        rev = 0
    try:
        imp_rev = imp_version_parts[2]
    except IndexError:
        imp_rev = 0

    if int(version_parts[0]) != int(imp_version_parts[0]):  # Major
        return False
    if int(version_parts[1]) > int(imp_version_parts[1]):  # Minor
        return False
    if (int(version_parts[1]) == int(imp_version_parts[1]) and
            int(rev) > int(imp_rev)):  # Revision
        return False
    return True


class DummyLock(object):
    def acquire(self):
        pass

    def release(self):
        pass

    def __enter__(self):
        self.acquire()

    def __exit__(self, type, value, traceback):
        self.release()




# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Notification drivers for sending notifications via messaging.

The messaging drivers publish notification messages to notification
listeners.

The driver will block the notifier's thread until the notification message has
been passed to the messaging transport.  There is no guarantee that the
notification message will be consumed by a notification listener.

Notification messages are sent 'at-most-once' - ensuring that they are not
duplicated.

If the connection to the messaging service is not active when a notification is
sent this driver will block waiting for the connection to complete.  If the
connection fails to complete, the driver will try to re-establish that
connection. By default this will continue indefinitely until the connection
completes. However, the retry parameter can be used to have the notification
send fail with a MessageDeliveryFailure after the given number of retries.
"""

import logging

import oslo_messaging
from oslo_messaging._i18n import _LE
from oslo_messaging.notify import notifier

LOG = logging.getLogger(__name__)


class MessagingDriver(notifier.Driver):

    """Send notifications using the 1.0 message format.

    This driver sends notifications over the configured messaging transport,
    but without any message envelope (also known as message format 1.0).

    This driver should only be used in cases where there are existing consumers
    deployed which do not support the 2.0 message format.
    """

    def __init__(self, conf, topics, transport, version=1.0):
        super(MessagingDriver, self).__init__(conf, topics, transport)
        self.version = version

    def notify(self, ctxt, message, priority, retry):
        priority = priority.lower()
        for topic in self.topics:
            target = oslo_messaging.Target(topic='%s.%s' % (topic, priority))
            try:
                self.transport._send_notification(target, ctxt, message,
                                                  version=self.version,
                                                  retry=retry)
            except Exception:
                LOG.exception(_LE("Could not send notification to %(topic)s. "
                                  "Payload=%(message)s"),
                              dict(topic=topic, message=message))


class MessagingV2Driver(MessagingDriver):

    "Send notifications using the 2.0 message format."

    def __init__(self, conf, **kwargs):
        super(MessagingV2Driver, self).__init__(conf, version=2.0, **kwargs)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging

import six

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging._i18n import _LW

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class Sender(object):

    @abc.abstractmethod
    def send_message(self, socket, multipart_message):
        """Send message to a socket from a multipart list."""


class CentralSender(Sender):

    def __init__(self):
        self._send_message_versions = \
            zmq_version.get_method_versions(self, 'send_message')

    def send_message(self, socket, multipart_message):
        message_version = multipart_message[zmq_names.MESSAGE_VERSION_IDX]
        if six.PY3:
            message_version = message_version.decode('utf-8')

        send_message_version = self._send_message_versions.get(message_version)
        if send_message_version is None:
            LOG.warning(_LW("Dropping message with unsupported version %s"),
                        message_version)
            return

        send_message_version(socket, multipart_message)


class LocalSender(Sender):
    pass


class CentralRouterSender(CentralSender):

    def _send_message_v_1_0(self, socket, multipart_message):
        message_type = int(multipart_message[zmq_names.MESSAGE_TYPE_IDX])
        routing_key = multipart_message[zmq_names.ROUTING_KEY_IDX]
        reply_id = multipart_message[zmq_names.REPLY_ID_IDX]
        message_id = multipart_message[zmq_names.MESSAGE_ID_IDX]
        message_version = multipart_message[zmq_names.MESSAGE_VERSION_IDX]

        socket.send(routing_key, zmq.SNDMORE)
        socket.send(b'', zmq.SNDMORE)
        socket.send(message_version, zmq.SNDMORE)
        socket.send(reply_id, zmq.SNDMORE)
        socket.send(multipart_message[zmq_names.MESSAGE_TYPE_IDX], zmq.SNDMORE)
        socket.send_multipart(multipart_message[zmq_names.MESSAGE_ID_IDX:])

        LOG.debug("Dispatching %(msg_type)s message %(msg_id)s - from %(rid)s "
                  "-> to %(rkey)s (v%(msg_version)s)",
                  {"msg_type": zmq_names.message_type_str(message_type),
                   "msg_id": message_id,
                   "rkey": routing_key,
                   "rid": reply_id,
                   "msg_version": message_version})


class CentralAckSender(CentralSender):

    def _send_message_v_1_0(self, socket, multipart_message):
        message_type = zmq_names.ACK_TYPE
        message_id = multipart_message[zmq_names.MESSAGE_ID_IDX]
        routing_key = socket.handle.identity
        reply_id = multipart_message[zmq_names.REPLY_ID_IDX]
        message_version = multipart_message[zmq_names.MESSAGE_VERSION_IDX]

        socket.send(reply_id, zmq.SNDMORE)
        socket.send(b'', zmq.SNDMORE)
        socket.send(message_version, zmq.SNDMORE)
        socket.send(routing_key, zmq.SNDMORE)
        socket.send(six.b(str(message_type)), zmq.SNDMORE)
        socket.send_string(message_id)

        LOG.debug("Sending %(msg_type)s for %(msg_id)s to %(rid)s "
                  "[from %(rkey)s] (v%(msg_version)s)",
                  {"msg_type": zmq_names.message_type_str(message_type),
                   "msg_id": message_id,
                   "rid": reply_id,
                   "rkey": routing_key,
                   "msg_version": message_version})


class CentralPublisherSender(CentralSender):

    def _send_message_v_1_0(self, socket, multipart_message):
        message_type = int(multipart_message[zmq_names.MESSAGE_TYPE_IDX])
        assert message_type in zmq_names.MULTISEND_TYPES, "Fanout expected!"
        topic_filter = multipart_message[zmq_names.ROUTING_KEY_IDX]
        message_id = multipart_message[zmq_names.MESSAGE_ID_IDX]
        message_version = multipart_message[zmq_names.MESSAGE_VERSION_IDX]

        socket.send(topic_filter, zmq.SNDMORE)
        socket.send(message_version, zmq.SNDMORE)
        socket.send(six.b(str(message_type)), zmq.SNDMORE)
        socket.send_multipart(multipart_message[zmq_names.MESSAGE_ID_IDX:])

        LOG.debug("Publishing message %(msg_id)s on [%(topic)s] "
                  "(v%(msg_version)s)",
                  {"topic": topic_filter,
                   "msg_id": message_id,
                   "msg_version": message_version})


class LocalPublisherSender(LocalSender):

    TOPIC_IDX = 0
    MSG_VERSION_IDX = 1
    MSG_TYPE_IDX = 2
    MSG_ID_IDX = 3

    def send_message(self, socket, multipart_message):
        socket.send_multipart(multipart_message)

        LOG.debug("Publishing message %(msg_id)s on [%(topic)s] "
                  "(v%(msg_version)s)",
                  {"topic": multipart_message[self.TOPIC_IDX],
                   "msg_id": multipart_message[self.MSG_ID_IDX],
                   "msg_version": multipart_message[self.MSG_VERSION_IDX]})



# Copyright 2013-2014 eNovance
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import webob

from oslo_messaging.notify import middleware
from oslo_messaging.tests import utils
from six.moves import mock


class FakeApp(object):
    def __call__(self, env, start_response):
        body = 'Some response'
        start_response('200 OK', [
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(sum(map(len, body))))
        ])
        return [body]


class FakeFailingApp(object):
    def __call__(self, env, start_response):
        raise Exception("It happens!")


class NotifierMiddlewareTest(utils.BaseTestCase):

    def test_notification(self):
        m = middleware.RequestNotifier(FakeApp())
        req = webob.Request.blank('/foo/bar',
                                  environ={'REQUEST_METHOD': 'GET',
                                           'HTTP_X_AUTH_TOKEN': uuid.uuid4()})
        with mock.patch(
                'oslo_messaging.notify.notifier.Notifier._notify') as notify:
            m(req)
            # Check first notification with only 'request'
            call_args = notify.call_args_list[0][0]
            self.assertEqual('http.request', call_args[1])
            self.assertEqual('INFO', call_args[3])
            self.assertEqual(set(['request']),
                             set(call_args[2].keys()))

            request = call_args[2]['request']
            self.assertEqual('/foo/bar', request['PATH_INFO'])
            self.assertEqual('GET', request['REQUEST_METHOD'])
            self.assertIn('HTTP_X_SERVICE_NAME', request)
            self.assertNotIn('HTTP_X_AUTH_TOKEN', request)
            self.assertFalse(any(map(lambda s: s.startswith('wsgi.'),
                                     request.keys())),
                             "WSGI fields are filtered out")

            # Check second notification with request + response
            call_args = notify.call_args_list[1][0]
            self.assertEqual('http.response', call_args[1])
            self.assertEqual('INFO', call_args[3])
            self.assertEqual(set(['request', 'response']),
                             set(call_args[2].keys()))

            request = call_args[2]['request']
            self.assertEqual('/foo/bar', request['PATH_INFO'])
            self.assertEqual('GET', request['REQUEST_METHOD'])
            self.assertIn('HTTP_X_SERVICE_NAME', request)
            self.assertNotIn('HTTP_X_AUTH_TOKEN', request)
            self.assertFalse(any(map(lambda s: s.startswith('wsgi.'),
                                     request.keys())),
                             "WSGI fields are filtered out")

            response = call_args[2]['response']
            self.assertEqual('200 OK', response['status'])
            self.assertEqual('13', response['headers']['content-length'])

    def test_notification_response_failure(self):
        m = middleware.RequestNotifier(FakeFailingApp())
        req = webob.Request.blank('/foo/bar',
                                  environ={'REQUEST_METHOD': 'GET',
                                           'HTTP_X_AUTH_TOKEN': uuid.uuid4()})
        with mock.patch(
                'oslo_messaging.notify.notifier.Notifier._notify') as notify:
            try:
                m(req)
                self.fail("Application exception has not been re-raised")
            except Exception:
                pass
            # Check first notification with only 'request'
            call_args = notify.call_args_list[0][0]
            self.assertEqual('http.request', call_args[1])
            self.assertEqual('INFO', call_args[3])
            self.assertEqual(set(['request']),
                             set(call_args[2].keys()))

            request = call_args[2]['request']
            self.assertEqual('/foo/bar', request['PATH_INFO'])
            self.assertEqual('GET', request['REQUEST_METHOD'])
            self.assertIn('HTTP_X_SERVICE_NAME', request)
            self.assertNotIn('HTTP_X_AUTH_TOKEN', request)
            self.assertFalse(any(map(lambda s: s.startswith('wsgi.'),
                                     request.keys())),
                             "WSGI fields are filtered out")

            # Check second notification with 'request' and 'exception'
            call_args = notify.call_args_list[1][0]
            self.assertEqual('http.response', call_args[1])
            self.assertEqual('INFO', call_args[3])
            self.assertEqual(set(['request', 'exception']),
                             set(call_args[2].keys()))

            request = call_args[2]['request']
            self.assertEqual('/foo/bar', request['PATH_INFO'])
            self.assertEqual('GET', request['REQUEST_METHOD'])
            self.assertIn('HTTP_X_SERVICE_NAME', request)
            self.assertNotIn('HTTP_X_AUTH_TOKEN', request)
            self.assertFalse(any(map(lambda s: s.startswith('wsgi.'),
                                     request.keys())),
                             "WSGI fields are filtered out")

            exception = call_args[2]['exception']
            self.assertIn('middleware.py', exception['traceback'][0])
            self.assertIn('It happens!', exception['traceback'][-1])
            self.assertEqual("Exception('It happens!',)", exception['value'])

    def test_process_request_fail(self):
        def notify_error(context, publisher_id, event_type,
                         priority, payload):
            raise Exception('error')
        with mock.patch('oslo_messaging.notify.notifier.Notifier._notify',
                        notify_error):
            m = middleware.RequestNotifier(FakeApp())
            req = webob.Request.blank('/foo/bar',
                                      environ={'REQUEST_METHOD': 'GET'})
            m.process_request(req)

    def test_process_response_fail(self):
        def notify_error(context, publisher_id, event_type,
                         priority, payload):
            raise Exception('error')
        with mock.patch('oslo_messaging.notify.notifier.Notifier._notify',
                        notify_error):
            m = middleware.RequestNotifier(FakeApp())
            req = webob.Request.blank('/foo/bar',
                                      environ={'REQUEST_METHOD': 'GET'})
            m.process_response(req, webob.response.Response())

    def test_ignore_req_opt(self):
        m = middleware.RequestNotifier(FakeApp(),
                                       ignore_req_list='get, PUT')
        req = webob.Request.blank('/skip/foo',
                                  environ={'REQUEST_METHOD': 'GET'})
        req1 = webob.Request.blank('/skip/foo',
                                   environ={'REQUEST_METHOD': 'PUT'})
        req2 = webob.Request.blank('/accept/foo',
                                   environ={'REQUEST_METHOD': 'POST'})
        with mock.patch(
                'oslo_messaging.notify.notifier.Notifier._notify') as notify:
            # Check GET request does not send notification
            m(req)
            m(req1)
            self.assertEqual(0, len(notify.call_args_list))

            # Check non-GET request does send notification
            m(req2)
            self.assertEqual(2, len(notify.call_args_list))
            call_args = notify.call_args_list[0][0]
            self.assertEqual('http.request', call_args[1])
            self.assertEqual('INFO', call_args[3])
            self.assertEqual(set(['request']),
                             set(call_args[2].keys()))

            request = call_args[2]['request']
            self.assertEqual('/accept/foo', request['PATH_INFO'])
            self.assertEqual('POST', request['REQUEST_METHOD'])

            call_args = notify.call_args_list[1][0]
            self.assertEqual('http.response', call_args[1])
            self.assertEqual('INFO', call_args[3])
            self.assertEqual(set(['request', 'response']),
                             set(call_args[2].keys()))



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging._drivers import common
from oslo_messaging._drivers.zmq_driver.client import zmq_client_base
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

zmq = zmq_async.import_zmq()


class WrongClientException(common.RPCException):
    """Raised if client type doesn't match configuration"""


class ZmqClientMixDirectPubSub(zmq_client_base.ZmqClientBase):
    """Client for using with direct connections and fanout over proxy:

        use_pub_sub = true
        use_router_proxy = false

    """

    def __init__(self, conf, matchmaker=None, allowed_remote_exmods=None):

        if conf.oslo_messaging_zmq.use_router_proxy or not \
                conf.oslo_messaging_zmq.use_pub_sub:
            raise WrongClientException()

        publisher_direct = self._create_publisher_direct(conf, matchmaker)
        publisher_proxy = self._create_publisher_proxy_dynamic(conf,
                                                               matchmaker)

        super(ZmqClientMixDirectPubSub, self).__init__(
            conf, matchmaker, allowed_remote_exmods,
            publishers={
                zmq_names.CAST_FANOUT_TYPE: publisher_proxy,
                zmq_names.NOTIFY_TYPE: publisher_proxy,
                "default": publisher_direct
            }
        )


class ZmqClientDirect(zmq_client_base.ZmqClientBase):
    """This kind of client (publishers combination) is to be used for
    direct connections only:

        use_pub_sub = false
        use_router_proxy = false
    """

    def __init__(self, conf, matchmaker=None, allowed_remote_exmods=None):

        if conf.oslo_messaging_zmq.use_pub_sub or \
                conf.oslo_messaging_zmq.use_router_proxy:
            raise WrongClientException()

        super(ZmqClientDirect, self).__init__(
            conf, matchmaker, allowed_remote_exmods,
            publishers={
                "default": self._create_publisher_direct(conf, matchmaker)
            }
        )


class ZmqClientProxy(zmq_client_base.ZmqClientBase):
    """Client for using with proxy:

        use_pub_sub = true
        use_router_proxy = true
    or
        use_pub_sub = false
        use_router_proxy = true
    """

    def __init__(self, conf, matchmaker=None, allowed_remote_exmods=None):

        if not conf.oslo_messaging_zmq.use_router_proxy:
            raise WrongClientException()

        super(ZmqClientProxy, self).__init__(
            conf, matchmaker, allowed_remote_exmods,
            publishers={
                "default": self._create_publisher_proxy(conf, matchmaker)
            }
        )



# Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_messaging.tests import utils as test_utils


class TestDeprecationWarning(test_utils.BaseTestCase):

    @mock.patch('warnings.warn')
    def test_impl_messaging_deprecation_warning(self, mock_warn):
        # Tests that we get a deprecation warning when loading a messaging
        # driver out of oslo_messaging.notify._impl_messaging.
        from oslo_messaging.notify import _impl_messaging as messaging
        driver = messaging.MessagingV2Driver(  # noqa
            conf={}, topics=['notifications'], transport='rpc')
        # Make sure we got a deprecation warning by loading from the alias
        self.assertEqual(1, mock_warn.call_count)



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client import zmq_response
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.server.consumers \
    import zmq_consumer_base
from oslo_messaging._drivers.zmq_driver.server import zmq_incoming_message
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging._i18n import _LE, _LI

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class RouterConsumer(zmq_consumer_base.SingleSocketConsumer):

    def __init__(self, conf, poller, server):
        self.reply_sender = zmq_senders.ReplySenderDirect(conf)
        super(RouterConsumer, self).__init__(conf, poller, server, zmq.ROUTER)
        self._receive_request_versions = \
            zmq_version.get_method_versions(self, 'receive_request')
        LOG.info(_LI("[%s] Run ROUTER consumer"), self.host)

    def _reply(self, rpc_message, reply, failure):
        if failure is not None:
            failure = rpc_common.serialize_remote_exception(failure)
        reply = zmq_response.Reply(message_id=rpc_message.message_id,
                                   reply_id=rpc_message.reply_id,
                                   message_version=rpc_message.message_version,
                                   reply_body=reply,
                                   failure=failure)
        self.reply_sender.send(rpc_message.socket, reply)
        return reply

    def _create_message(self, context, message, message_version, reply_id,
                        message_id, socket, message_type):
        if message_type == zmq_names.CALL_TYPE:
            message = zmq_incoming_message.ZmqIncomingMessage(
                context, message, message_version=message_version,
                reply_id=reply_id, message_id=message_id,
                socket=socket, reply_method=self._reply
            )
        else:
            message = zmq_incoming_message.ZmqIncomingMessage(context, message)

        LOG.debug("[%(host)s] Received %(msg_type)s message %(msg_id)s "
                  "(v%(msg_version)s)",
                  {"host": self.host,
                   "msg_type": zmq_names.message_type_str(message_type),
                   "msg_id": message_id,
                   "msg_version": message_version})
        return message

    def _get_receive_request_version(self, version):
        receive_request_version = self._receive_request_versions.get(version)
        if receive_request_version is None:
            raise zmq_version.UnsupportedMessageVersionError(version)
        return receive_request_version

    def receive_request(self, socket):
        try:
            reply_id = socket.recv()
            assert reply_id != b'', "Valid reply id expected!"
            empty = socket.recv()
            assert empty == b'', "Empty delimiter expected!"
            message_version = socket.recv_string()
            assert message_version != b'', "Valid message version expected!"

            receive_request_version = \
                self._get_receive_request_version(message_version)
            return receive_request_version(reply_id, socket)
        except (zmq.ZMQError, AssertionError, ValueError,
                zmq_version.UnsupportedMessageVersionError) as e:
            LOG.error(_LE("Receiving message failed: %s"), str(e))
            # NOTE(gdavoian): drop the left parts of a broken message
            if socket.getsockopt(zmq.RCVMORE):
                socket.recv_multipart()

    def _receive_request_v_1_0(self, reply_id, socket):
        message_type = int(socket.recv())
        assert message_type in zmq_names.REQUEST_TYPES, "Request expected!"
        message_id = socket.recv_string()
        assert message_id != '', "Valid message id expected!"
        context, message = socket.recv_loaded()

        return self._create_message(context, message, '1.0', reply_id,
                                    message_id, socket, message_type)

    def cleanup(self):
        LOG.info(_LI("[%s] Destroy ROUTER consumer"), self.host)
        super(RouterConsumer, self).cleanup()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._i18n import _


# current driver's version for representing internal message format
MESSAGE_VERSION = '1.0'


class UnsupportedMessageVersionError(rpc_common.RPCException):
    msg_fmt = _("Message version %(version)s is not supported.")

    def __init__(self, version):
        super(UnsupportedMessageVersionError, self).__init__(version=version)


def get_method_versions(obj, method_name):
    """Useful function for initializing versioned senders/receivers.

    Returns a dictionary of different internal versions of the given method.

    Assumes that the object has the particular versioned method and this method
    is public. Thus versions are private implementations of the method.

    For example, for a method 'func' methods '_func_v_1_0', '_func_v_1_5',
    '_func_v_2_0', etc. are assumed as its respective 1.0, 1.5, 2.0 versions.
    """

    assert callable(getattr(obj, method_name, None)), \
        "Object must have specified method!"
    assert not method_name.startswith('_'), "Method must be public!"

    method_versions = {}
    for attr_name in dir(obj):
        if attr_name == method_name:
            continue
        attr = getattr(obj, attr_name, None)
        if not callable(attr):
            continue
        match_obj = re.match(r'^_%s_v_(\d)_(\d)$' % method_name, attr_name)
        if match_obj is not None:
            version = '.'.join([match_obj.group(1), match_obj.group(2)])
            method_versions[version] = attr

    return method_versions



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging._drivers.zmq_driver.proxy import zmq_sender
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_socket


zmq = zmq_async.import_zmq()


class PublisherProxy(object):
    """PUB/SUB based request publisher

        The publisher intended to be used for Fanout and Notify
        multi-sending patterns.

        It differs from direct publishers like DEALER or PUSH based
        in a way it treats matchmaker. Here all publishers register
        in the matchmaker. Subscribers (server-side) take the list
        of publishers and connect to all of them but subscribe
        only to a specific topic-filtering tag generated from the
        Target object.
    """

    def __init__(self, conf, matchmaker, sender=None):
        super(PublisherProxy, self).__init__()
        self.conf = conf
        self.zmq_context = zmq.Context()
        self.matchmaker = matchmaker

        port = conf.zmq_proxy_opts.publisher_port
        self.socket = zmq_socket.ZmqFixedPortSocket(
            self.conf, self.zmq_context, zmq.PUB, conf.zmq_proxy_opts.host,
            port) if port != 0 else \
            zmq_socket.ZmqRandomPortSocket(
                self.conf, self.zmq_context, zmq.PUB, conf.zmq_proxy_opts.host)

        self.host = self.socket.connect_address
        self.sender = sender or zmq_sender.CentralPublisherSender()

    def send_request(self, multipart_message):
        self.sender.send_message(self.socket, multipart_message)

    def cleanup(self):
        self.socket.close()
